#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
This file contains any constants shared in the data model that
aren't more appropriately held in one of the object-focused files
"""

NO_PROJECT_NAME = "NO_PROJECT"
