/*
 * Copyright (c) Facebook, Inc. and its affiliates.
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */

.container {
  padding: 0px 30px;
  display: flex;
  max-width: 1280px;
  margin: 50px auto 0;
  align-items: flex-start;
}

.container > div {
  flex-grow: 1;
  margin-right: 20px;
  max-width: 33%;
}

.container > div:last-child {
  margin-right: 0px;
}

.above-the-fold {
  background-repeat: no-repeat;
  background-position: bottom;
  height: 458px;
  z-index: -1;
  /* background-image: url(https://www.rev.com/content/curve/img/home/bg-home-background.svg); */
  background-image: url(./Oval.svg);
  position: absolute;
  width: 100%;
  background-size: cover !important;
}

header {
  padding: 30px;
  max-width: 1280px;
  margin: 0 auto;
}

h1 {
  margin: 0;
  font-size: 40px;
}

.step-badge {
  margin-right: 10px;
}

.widget {
  animation: fadeIn 0.3s cubic-bezier(0.5, 0, 0, 1);
  animation-fill-mode: forwards;
  opacity: 0;
  border-radius: 10px !important;
  will-change: transform;
}

.widget:nth-child(1) {
  animation-delay: 0.1s;
}
.widget:nth-child(2) {
  animation-delay: 0.25s;
}
.widget:nth-child(3) {
  animation-delay: 0.4s;
}
.widget:nth-child(4) {
  animation-delay: 0.55s;
}

@keyframes fadeIn {
  0% {
    opacity: 0;
    transform: scale(0.9, 0.9) translateY(80px);
  }
  60% {
    transform: translateY(-2px);
  }
  100% {
    opacity: 1;
    transform: scale(1, 1) translateY(0px);
  }
}

.widget .bullet {
  margin-bottom: 5px;
}

.params-list {
  overflow-y: auto;
  overflow-x: none;
  max-height: 50px;
  color: #757677;
  font-style: italic;
  border-radius: 3px;
  box-shadow: inset 0 0 0 1px rgba(16, 22, 26, 0.2);
  background: rgba(255, 255, 255, 0.7);
  padding: 2px 5px;
}

.params-list .param-name {
}

h1 a,
h1 a:hover {
  color: inherit;
}

.params-list .param-value {
  color: #333;
}

.details {
  font-size: 13px;
  margin-top: 10px;
}

.metrics {
  display: flex;
  flex-direction: row;
  justify-content: center;
}

.metric {
  margin: 0px 0px 0px;
  text-align: center;
  padding: 10px;
  font-size: 16px;
  color: #333;
  flex: 1;
}

.metrics.highlight-first .metric:first-child {
  font-weight: bold;
}

.metric label {
  display: block;
  font-size: 12px;
}

.run-header {
  background-color: rgb(245, 248, 250);
  padding: 10px;
  margin-bottom: 10px;
  border-color: rgb(191, 204, 214);
  border-width: 1px;
  border-style: solid;
  border-radius: 5px;
}

.run-header.interactive:hover {
  background-color: #ebf1f5;
  transform: scale(1.02, 1.02);
  transition: transform 0.1s ease-out;
}
