/*
 * Copyright (c) Facebook, Inc. and its affiliates.
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */
export function pluralize(num: Number, word: string, plural?: string): string {
  if (num === 1) {
    return word;
  } else return plural || word + "s";
}
