#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


class ParlAIAgent:
    def __init__(self, *args, **kwargs):
        raise NotImplementedError("You need to install ParlAI to use this blueprint")


class Message:
    def __init__(self, *args, **kwargs):
        raise NotImplementedError("You need to install ParlAI to use this blueprint")
