<html style="font-size: 16px">
  <head>
    <title>Task content missing!</title>
  </head>
  <body>
    <section style="position: relative; display: block">
      <div
        style="
          justify-content: center;
          display: flex !important;
          flex-direction: column;
          position: relative;
        "
      >
        <h2 style="margin: 30px auto 0">
          Oops! Something went wrong with this task...
        </h2>
        <p style="margin: 30px auto 0; max-width: 650px">
          If you're seeing this page, it's because the server that was hosting
          the task that you were trying to reach is no longer accessible. Often
          this happens when the author of the task requests it to be taken down
          but the request failed, or if something goes severely wrong with the
          task setup.
        </p>
        <p style="margin: 30px auto 0; max-width: 650px">
          If you were intending on seeing a task here, we've logged and notified
          the task owner so that this outstanding task can be taken down. Sorry
          for the inconvenience, and thank you for your work!
        </p>
      </div>
    </section>
  </body>
</html>
