#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


"""
.. include:: README.md
"""
__docformat__ = "restructuredtext"

__all__ = [
    "architect",
    "blueprint",
    "crowd_provider",
    "database",
    "architects",
    "blueprints",
    "providers",
    "databases",
]
